/*
 *
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Swapnil V. Patil
 * Copyright (c) 2003-2004 Anand Kashyap
 * Copyright (c) 2003-2004 Gopalan Sivathanu
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */


#ifdef HAVE_CONFIG_H
# include <config.h>
# include <linux/fs.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_i3fs.h"
#endif /* FISTGEN */
#include "i3fs_utils.h"
#include "i3fs.h"


# include <unistd.h>

/* write a message to the logfile specified during mount */

void write_to_logfile(super_block_t *sb, char *message)	{
	struct file *fp;
	long long int a;
	mm_segment_t oldfs;

	oldfs = get_fs();
	set_fs(KERNEL_DS);

	fp = filp_open(stopd(sb)->logfile, O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
	a = fp->f_op->write(fp, message, strlen(message), &fp->f_pos);

	set_fs(oldfs);
	filp_close(fp, NULL);
	
}

/* Write failure message to syslog */

void write_to_syslog(char *message)	{
	printk(KERN_WARNING "%s", message);
}


/* Create message to be logged when there is a metadata checksum mismatch */

void report_metadata_mismatch(struct inode *inode, struct task_struct* process)	{
	char *message, *tmp;
	struct timeval tv;
	

	do_gettimeofday(&tv);
	message = KMALLOC(4096, GFP_KERNEL);
	tmp = KMALLOC(1024, GFP_KERNEL);

	strcpy(message, "\nMETADATA CHECKSUM MISMATCH");
	strcat(message, "\n");
	
	sprintf(tmp, "Timestamp: %ld\n", tv.tv_sec);
	strcat(message, tmp);
	sprintf(tmp, "Inode number: %lu \n", itohi(inode)->i_ino);
	strcat(message, tmp);
	sprintf(tmp, "Process ID: %d \n", process->pid);
	strcat(message, tmp);
	sprintf(tmp, "Process Name: %s\n", process->comm);
	strcat(message, tmp);
	sprintf(tmp, "Effective User ID: %u \n", process->euid);
	strcat(message, tmp);
	sprintf(tmp, "Effective Group ID: %u \n", process->egid);
	strcat(message, tmp);

	if (stopd(inode->i_sb)->logfile) 
		write_to_logfile(inode->i_sb, message);
	else
		write_to_syslog(message);

	KFREE(message);
	KFREE(tmp);

}

/* Create message to be logged when there is a data checksum mismatch */

void report_data_mismatch(struct dentry *dentry, struct task_struct* process) {
	char *message, *tmp;
	struct timeval tv;


        do_gettimeofday(&tv);


        message = KMALLOC(4096, GFP_KERNEL);
        tmp = KMALLOC(1024, GFP_KERNEL);

        strcpy(message, "\nDATA CHECKSUM MISMATCH");
        strcat(message, "\n");
	
	sprintf(tmp, "Timestamp: %ld\n", tv.tv_sec);
        strcat(message, tmp);
	sprintf(tmp, "Filename: %s \n", dentry->d_name.name);
        strcat(message, tmp);
        sprintf(tmp, "Inode number: %lu \n", itohi(dentry->d_inode)->i_ino);
        strcat(message, tmp);
        sprintf(tmp, "Process ID: %d \n", process->pid);
        strcat(message, tmp);
        sprintf(tmp, "Process Name: %s\n", process->comm);
        strcat(message, tmp);
        sprintf(tmp, "Effective User ID: %u \n", process->euid);
        strcat(message, tmp);
        sprintf(tmp, "Effective Group ID: %u \n", process->egid);
        strcat(message, tmp);

        if (stopd(dentry->d_inode->i_sb)->logfile)
                write_to_logfile(dentry->d_inode->i_sb, message);
        else
                write_to_syslog(message);

}


/* Verify the passphrase provided during mount */


int verify_passphrase(DB *policydb, void *checksum)	{
	DBT data, key;
        int ret=0, err=0;
        
        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number

        key.size = 9;
        key.data = "authtoken";
	ret = policydb->get(policydb, NULL, &key, &data, 0);

	if (ret != 0)	{
		data.data = checksum;
		data.size = CHECKSUM_SIZE;
		
		ret = policydb->put(policydb, NULL, &key, &data, 0);
		if (ret != 0)	{
			printk(KERN_WARNING "Cannot insert auth token");
			return 0;
		} else	{
			printk("\nAuth token successfully inserted.");
		}
		policydb->sync(policydb, 0);
		return 1;
	}


	if (memcmp(checksum, data.data, CHECKSUM_SIZE) != 0)	{
		printk(KERN_WARNING "\nInvalid passphrase entered");
		return 0;
	} else	{
		return 1;
	}

}


/* For debugging purposes -- function to print a void pointer data */

void printvoid(void *data, int len, int flag)	{
	int i=0;
	printk("\n");
	for (i=0; i<len; i++)	{
		if (flag)
			printk("%d,",*(char *)(data+i));
		else 
			printk("%c", *(char *)(data+i));
	}
		
};


/* Generate hmac md5 checksums for the entire file */


void hmac_file(struct file* fp, const unsigned char *key, int key_len, void *digest)
/* unsigned char*  text;              * pointer to data stream 
 * int             text_len;          * length of data stream
 * unsigned char*  key;               * pointer to authentication key
 * int             key_len;           * length of authentication key
 * caddr_t         digest;            * caller digest to be filled in
 */
{
        void  *context;
        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
        unsigned char *tk = NULL;
	long int size;
	long long offset=0, bytes_read=1;
	char *text;
	mm_segment_t oldfs;
        int i;

        /* if key is longer than 64 bytes reset it to key=MD5(key) */

	text = KMALLOC(PAGE_SIZE, GFP_KERNEL);

        if (key_len > 64) {
                tk = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
                if (!tk) {
                        goto out;

                }
                if (md5_initctx(&context) != 0) {
                        goto out;
                }

                md5_append(context, key, key_len);
                md5_finish(context, tk);
                md5_freectx(context);
                context = NULL;

                key = tk;
                key_len = CHECKSUM_SIZE;
        }

        /*
         * the HMAC_MD5 transform looks like:
         *
         * MD5(K XOR opad, MD5(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times
         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        memset( k_ipad, '\0', sizeof k_ipad);
        memset( k_opad, '\0', sizeof k_opad);
        memcpy( k_ipad, key, key_len);
        memcpy( k_opad, key, key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }


        /*
         * perform inner MD5
         */
        md5_initctx(&context);                   /* init context for 1st
                                              * pass */
        md5_append(context, k_ipad, 64);      /* start with inner pad */

	size = fp->f_dentry->d_inode->i_size;
	
	while (bytes_read > 0)	{
	
		oldfs = get_fs();
                set_fs(KERNEL_DS);

		bytes_read = fp->f_op->read(fp, text, PAGE_SIZE, &fp->f_pos);
	        md5_append(context, text, bytes_read); /* then text of datagram */

		set_fs(oldfs);

	}

        md5_finish(context, digest);          /* finish up 1st pass */
        md5_freectx(context);
        context = NULL;

        /*
         * perform outer MD5
         */
        md5_initctx(&context);                   /* init context for 2nd
                                              * pass */
        md5_append(context, k_opad, 64);     /* start with outer pad */
        md5_append(context, digest, 16);     /* then results of 1st
                                              * hash */
        md5_finish(context, digest);          /* finish up 2nd pass */
        md5_freectx(context);
        context = NULL;

out:
	if (text)	{
		KFREE(text);
	}
        if (context) {
                md5_freectx(context);
        }
        if (tk) {
                KFREE(tk);
        }
}


/* Generate hmac md5 checksum for arbitrary sized data */


void hmac(const unsigned char *text, int text_len, const unsigned char *key, int key_len, void *digest)
/* unsigned char*  text;              * pointer to data stream 
 * int             text_len;          * length of data stream
 * unsigned char*  key;               * pointer to authentication key
 * int             key_len;           * length of authentication key
 * caddr_t         digest;            * caller digest to be filled in
 */
{
        void  *context;
        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
        unsigned char *tk = NULL;
        int i;

        /* if key is longer than 64 bytes reset it to key=MD5(key) */


	if (key_len > 64) {
                tk = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
                if (!tk) {
                        goto out;

                }
                if (md5_initctx(&context) != 0) {
                        goto out;
                }

                md5_append(context, key, key_len);
                md5_finish(context, tk);
                md5_freectx(context);
                context = NULL;

                key = tk;
                key_len = CHECKSUM_SIZE;
        }

        /*
         * the HMAC_MD5 transform looks like:
         *
         * MD5(K XOR opad, MD5(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times
         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        memset( k_ipad, '\0', sizeof k_ipad);
        memset( k_opad, '\0', sizeof k_opad);
        memcpy( k_ipad, key, key_len);
        memcpy( k_opad, key, key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }


        /*
         * perform inner MD5
         */
        md5_initctx(&context);                   /* init context for 1st
                                              * pass */
        md5_append(context, k_ipad, 64);      /* start with inner pad */
        md5_append(context, text, text_len); /* then text of datagram */
        md5_finish(context, digest);          /* finish up 1st pass */
        md5_freectx(context);
        context = NULL;

        /*
         * perform outer MD5
         */
        md5_initctx(&context);                   /* init context for 2nd
                                              * pass */
        md5_append(context, k_opad, 64);     /* start with outer pad */
        md5_append(context, digest, 16);     /* then results of 1st
                                              * hash */
        md5_finish(context, digest);          /* finish up 2nd pass */
        md5_freectx(context);
        context = NULL;

out:
        if (context) {
                md5_freectx(context);
        }
        if (tk) {
                KFREE(tk);
        }
}


/* Check if policy exists for a given inode number
 * bits is passed by reference and if a policy exists
 * then 'bits' will be populated by the policy bits
 */


int
policy_exists(DB *policydb, long int ino, int *bits, int *freq)	{
	DBT data, key;
	int ret=0, err=0;
	
	memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number

        key.size = sizeof(long int);
        key.data = KMALLOC(sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }

	*(long int*)(key.data) = ino;
	
	
	ret = policydb->get(policydb, NULL, &key, &data, 0);
	if (ret == 0)	{
		*bits = *((int*)data.data);
		*freq = *((int*)(data.data + 4));
		KFREE(key.data);
		return 1;
	}
	
        KFREE(key.data);
out:
	return 0;
}  

/* Given an inode number, this function gives the meta-data checksum 
 * for this inode.  Function should be called after verifying that a policy
 * exists for the meta-data.
 */


int
get_from_metadb(DB* metadb, long int ino, void *checksum)	{
        DBT data, key;
        int ret=0, err=0;

        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number

        key.size = sizeof(long int);
        key.data = KMALLOC(sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }

        *(long int*)(key.data) = ino;

	ret = metadb->get(metadb, NULL, &key, &data, 0);
        if (ret == 0)   {
		memcpy(checksum, data.data, data.size);
                KFREE(key.data);
                return 1;
        }
	
        KFREE(key.data);
out:
	return 0;
}


/* Just inserts the checksum to the data database */

int add_to_datadb(DB *datadb, long int ino, long int pno, void *checksum)       {
        DBT data, key;
        int err=1;

        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number and the page number

        key.size = 2 * sizeof(long int);
        key.data = KMALLOC(2 * sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }

        *(long int*)(key.data) = ino;
        *(long int*)(key.data + 4) = pno;

        // data is the checksum
        data.size = CHECKSUM_SIZE;
        data.data = checksum;

        // insert entry to the database

        if (datadb->put(datadb, NULL, &key, &data, 0) != 0)
                err = -EIO;

        KFREE(key.data);

out:
        return err;
}


/* Inserts data to the access counter database */

int add_to_accessdb(DB *accessdb, long int ino)       {
        DBT data, key;
        int err=1;

        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number and the page number

        key.size = sizeof(long int);
        key.data = KMALLOC(2 * sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }

        *(long int*)(key.data) = ino;

        // data is the checksum
        data.size = sizeof(long int);
        data.data = KMALLOC(sizeof(long int), GFP_KERNEL);
	
	*(long int*)data.data = 0;

        // insert entry to the database

        if (accessdb->put(accessdb, NULL, &key, &data, 0) != 0)
                err = -EIO;

        KFREE(key.data);
	KFREE(data.data);

out:
        return err;
}



/* Wrapper for computing checksum for arbitrary length data */

void compute_cksum(void *buf, int size, void *key, int keylen, void *checksum)   {
	hmac(buf, size, key, keylen, checksum);
	return;
}

/* gets a filename and computes checksums for each of its pages
        and stores them in the datadb */

int checksum_filedata_per_page(dentry_t *dentry, file_t *fp, struct vfsmount *mnt)   {
        struct file  *fpi;
        struct super_block *sb;
        struct inode *inode, *hidden_inode;
        void *buf, *checksum;
        int err=1, ret,  i;
	long long offset=0, bytes_read=0, size=0;
	unsigned long page_number=0;
	mm_segment_t oldfs;
        DB *datadb;

	inode = dentry->d_inode;
	hidden_inode = itohi(inode);	

	if (S_ISDIR(inode->i_mode) || !(S_ISREG(inode->i_mode)))
                return 1;


	if (!(fp->f_op->read))
                return 1;


        buf = KMALLOC(PAGE_SIZE, GFP_KERNEL);
        if (!buf)       {
                err = -ENOMEM;
                goto out;
        }

        checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
        memset(checksum, 0, CHECKSUM_SIZE);


        if (!checksum)  {
                KFREE(buf);
                err = -ENOMEM;
                goto out;
        }

	
        sb = inode->i_sb;
        datadb = stopd(sb)->datadb;
        size = hidden_inode->i_size;

        // for each page, compute checksums and add to the database.

	bytes_read = 1;

	while (bytes_read > 0)	{
                memset(buf, 0, PAGE_SIZE);
                memset(checksum, 0, CHECKSUM_SIZE);

		oldfs = get_fs();
                set_fs(KERNEL_DS);

                bytes_read = fp->f_op->read(fp, buf, PAGE_SIZE, &fp->f_pos);

		offset = offset + bytes_read;

                compute_cksum(buf, bytes_read, stopd(sb)->passphrase, stopd(sb)->plen, checksum);
		
		set_fs(oldfs);
				

                if (!add_to_datadb(datadb, hidden_inode->i_ino, page_number, checksum))        		{       
			err = -EIO;
                        goto kout;
                }
		page_number++;
	
        }
kout:
        KFREE(checksum);
        KFREE(buf);
	
out:

	
        return err;
}


/* Function to checksum a file in its entirety and write to database*/

int checksum_filedata_to_db(dentry_t *dentry, struct file *fp)	{
	struct file *hidden_file;
        struct super_block *sb;
        struct inode *inode, *hidden_inode;
        void *buf, *checksum;
        int offset=0, size, err=1, ret, bytes_read=0, page_number=0;
        DB *datadb;


	inode = dentry->d_inode;
	hidden_inode = itohi(inode);


        buf = KMALLOC(PAGE_SIZE, GFP_KERNEL);
        if (!buf)       {
                err = -ENOMEM;
                goto out;
        }

        checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
	memset(checksum, 0, CHECKSUM_SIZE);


        if (!checksum)  {
                KFREE(buf);
                err = -ENOMEM;
                goto out;
        }


        sb = inode->i_sb;
        datadb = stopd(sb)->datadb;
        size = hidden_inode->i_size;

	

	hmac_file(fp, stopd(sb)->passphrase, stopd(sb)->plen, checksum);

	if (!add_to_datadb(datadb, hidden_inode->i_ino, 0, checksum))	{
		err = -EIO;
		KFREE(buf);
	        KFREE(checksum);
		goto out;
	}
	
	
	KFREE(buf);
	KFREE(checksum);

out:
	return err;
}

/*  get the stored checksum for the page from the data db
*/

int
get_stored_data_checksum(DB* datadb, long int ino, unsigned long pno, void *checksum)   {
        DBT data, key;
        int ret=0, err=0;

        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // key here is the inode number and the page number

        key.size = 2 * sizeof(long int);
        key.data = KMALLOC(2 * sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }


	*((long int*)(key.data)) = ino;
        *((long int*)(key.data + 4)) = pno;


        // get from the data database.

        if (datadb->get(datadb, NULL, &key, &data, 0) != 0)
                ret = 0;
        else	{
		memcpy(checksum, data.data, data.size);
                ret = 1;
	}
out:
	KFREE(key.data);
        return ret;
}


/* Compute file data checksum */

void compute_filedata_checksum(struct file* fp, super_block_t *sb, char *checksum)	{

	hmac_file(fp, stopd(sb)->passphrase, stopd(sb)->plen, checksum);
	return;
	
}

/* add checksum to metadata database */
	
int add_to_metadb(DB *metadb, long int ino, void *checksum)     {
        DBT data, key;
        int err=1;

        memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));

        // here key is the inode number

        key.size = sizeof(long int);
        key.data = KMALLOC(sizeof(long int), GFP_KERNEL);

        if (!key.data)  {
                err = -ENOMEM;
                goto out;
        }

        *(long int*)(key.data) = ino;

        // data is the checksum

        data.size = CHECKSUM_SIZE;
        data.data = checksum;

        if (metadb->put(metadb, NULL, &key, &data, 0) != 0)
                err = -EIO;

        KFREE(key.data);

out:
        return err;

}

/* Check the fields of the inode to be checksummed and then compute checksum
 * and write to the metadata checksum database */

int checksum_metadata(inode_t *i3fs_inode, int bits) {
        DB *metadb;
        struct meta_data mdata;
        void *checksum;
        int err, ret;
        struct inode *inode=NULL;


        metadb = stopd(i3fs_inode->i_sb)->metadb;
        inode = itohi(i3fs_inode);

	memset(&mdata, 0, sizeof(mdata));

        // Based on policy bits assign values -- if a field need not be checksummed,
        // assign it to 0.

        if (bits & PERMS) mdata.perm = inode->i_mode; else mdata.perm = 0;
        if (bits & INO) mdata.ino = inode->i_ino; else mdata.ino = 0;
        if (bits & NUM_LINKS) mdata.num_links = inode->i_nlink; else mdata.num_links = 0;  
        if (bits & UID) mdata.uid = inode->i_uid; else mdata.uid = 0;
        if (bits & GID) mdata.gid = inode->i_gid; else mdata.gid = 0;
        if (bits & SIZE) mdata.size = inode->i_size; else mdata.size = 0;
        if (bits & DEV_ID) mdata.dev_id = inode->i_dev; else mdata.dev_id = 0;
        if (bits & NUM_BLOCKS) mdata.num_blocks = inode->i_blocks; else mdata.num_blocks = 0;
        if (bits & A_TIME) mdata.atime = inode->i_atime; else mdata.atime = 0;
        if (bits & M_TIME) mdata.mtime = inode->i_mtime; else mdata.mtime = 0;
        if (bits & C_TIME) mdata.ctime = inode->i_ctime; else mdata.ctime = 0;

        checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);


        if (!checksum)  {
                err = -ENOMEM;
                goto out;
        }
	memset(checksum, 0, CHECKSUM_SIZE);

        // compute the checksum and add it to meta data database

        compute_cksum(&mdata, sizeof(mdata), stopd(i3fs_inode->i_sb)->passphrase, stopd(i3fs_inode->i_sb)->plen, checksum);



        err = add_to_metadb(metadb, inode->i_ino, checksum);
        KFREE(checksum);

out:
        return err;
}



void insert_to_accessdb(DB *accessdb, struct inode *inode)	{
	if (!add_to_accessdb(accessdb, inode->i_ino))	{
		printk("\nAdd to access DB failed");
	}
}
	


/* write policy it into the On-disk database */
void put_policy_in_database(struct super_block *sb, struct policy *policy) {
	struct nameidata nd;
	char *write_buffer=NULL, *bits;
	int write_size, err=0, ret=0;
	struct inode *inode, *i3fs_inode;
	DBT key, data;
	DB *policydb, *accessdb;
	dentry_t *dentry=NULL;
	struct file *fp=NULL;

	policydb = stopd(sb)->policydb;
	accessdb = stopd(sb)->accessdb;

	memset(&data, 0, sizeof(data));
	memset(&key, 0, sizeof(key));

	key.data = KMALLOC(sizeof(long int), GFP_KERNEL);
	data.data = KMALLOC(2 * sizeof(long int), GFP_KERNEL);
	key.size = sizeof(long int);
	data.size = 2 * sizeof(long int);
	
// Convert pathname to inode number	

	
	if ((err = path_init(policy->file, LOOKUP_FOLLOW, &nd)))
		ret = path_walk(policy->file, &nd);


	if (!S_ISREG((nd.dentry)->d_inode->i_mode) && !S_ISDIR((nd.dentry)->d_inode->i_mode))	{
		path_release(&nd);
		KFREE(key.data);
		KFREE(data.data);
		return;
	}


	i3fs_inode = (nd.dentry)->d_inode;
	

	if ( !(inode = itohi(i3fs_inode)) )	{
		printk(KERN_WARNING "\nHidden inode not present.  File doesn't belong to this file system");
		return;
	}
		
	*((long int*)(key.data)) = inode->i_ino;

	
	*((int*)data.data) = policy->policy_bits;
	*((int*)(data.data + 4)) = policy->freq_number;


	
	data.size = 2 * sizeof(int);

	if (policy->freq_number != 0)
		insert_to_accessdb(accessdb, inode);
	

	checksum_metadata((nd.dentry)->d_inode, policy->policy_bits);

	if (policy->policy_bits & DATA)	{
		if (policy->policy_bits & PER_PAGE)	{
				fp = filp_open(policy->file, O_RDONLY, 0);
				checksum_filedata_per_page(nd.dentry, ftohf(fp), nd.mnt);
				filp_close(fp, NULL);
		}
			else	{
				fp = filp_open(policy->file, O_RDONLY, 0);
				checksum_filedata_to_db(nd.dentry, ftohf(fp));
				filp_close(fp, NULL);
		}
	}

        err = policydb->put(policydb, NULL, &key, &data, 0);
        if (err != 0)   {
                printk("\nPutting to db failed!!!");
        }
//	printk("\nSuccess: Policy put to database.");
	dput(dentry);
	KFREE(key.data);
        KFREE(data.data);
	path_release(&nd);


}

//insert policy for new file


void insert_policy_in_database(inode_t *dir, dentry_t *dentry, int bits, int freq) {
        char *write_buffer=NULL;
        int write_size, err=0, ret=0;
        struct inode *inode=NULL, *hidden_inode;
        DBT key, data;
        DB *policydb;
	struct file *fp;


	//printk("\nInheritting policy.");


        policydb = stopd(dentry->d_inode->i_sb)->policydb;

	inode = dentry->d_inode;
	hidden_inode = itohi(inode);

        memset(&data, 0, sizeof(data));
        memset(&key, 0, sizeof(key));

        key.data = KMALLOC(sizeof(long int), GFP_KERNEL);
        data.data = KMALLOC(2 * sizeof(long int), GFP_KERNEL);
        key.size = sizeof(long int);
        data.size = 2 * sizeof(long int);


        *((long int*)(key.data)) = hidden_inode->i_ino;


        *((int*)data.data) = bits;
        *(int*)(data.data + 4) = freq;

        data.size = sizeof(int);


        checksum_metadata(dentry->d_inode, bits);
        err = policydb->put(policydb, NULL, &key, &data, 0);
        if (err != 0)   {
                printk("\nPutting to db failed!!!");
        }
	KFREE(key.data);
	KFREE(data.data);

}


/* Removes a record from any db given the key */

int remove_from_db(DB* db, void* kdata , int ksize)	{
	DBT key;

	key.data = kdata;
	key.size = ksize;

	if ((db->del(db, NULL, &key, 0)) != 0)	{
		printk(KERN_WARNING "Error while deleting from database");
		return -ENOENT;
	}
	return 1;
}


/* Removes a policy and all associated checksums from the databases.
 * Called from the REMOVE_POLICY ioctl. */


int remove_policy(char *pathname)	{
	struct nameidata nd;
        char *write_buffer=NULL;
        int write_size, err=0, ret=0, bits=0, freq=0;
        struct inode *inode, *hidden_inode;
	struct super_block *sb;
	int i;
        DB *policydb, *datadb, *metadb, *accessdb;
	void *key;

        if ((err = path_init(pathname, LOOKUP_FOLLOW, &nd)))
                ret = path_walk(pathname, &nd);

	inode = (nd.dentry)->d_inode;
	hidden_inode = itohi(inode);
	sb = inode->i_sb;


	policydb = stopd(sb)->policydb;
	datadb = stopd(sb)->datadb;
	metadb = stopd(sb)->metadb;
	accessdb = stopd(sb)->accessdb;

	if (!policy_exists(policydb, hidden_inode->i_ino, &bits, &freq))
		return -ENOENT;

	key = KMALLOC(2 * sizeof(long int), GFP_KERNEL);


	if (bits & DATA)	{
		if (bits & PER_PAGE)	{
			for (i=0; i<hidden_inode->i_size/PAGE_SIZE; i++)	{
				*(long int*)key = hidden_inode->i_ino;
				*(long int*)(key+sizeof(long int)) = i;
				remove_from_db(datadb, key, 8);
			}
		} else	{
			*(long int*)key = hidden_inode->i_ino;
                        *(long int*)(key+sizeof(long int)) = 0;
			remove_from_db(datadb, key, 8);
		}
	}


	path_release(&nd);

	*(long int*)key = hidden_inode->i_ino;
	remove_from_db(metadb, key, 4);
	remove_from_db(accessdb, key, 4);
	remove_from_db(policydb, key, 4);

	KFREE(key);

	return 1;

}


/* Check whether the inode has a frequency of checks condition associated with it
 * and if yes, check if the number of accesses has exceeded the condition.  If 
 * integrity check is required, returns 1 if not returns 0.
 */

int
frequency_pass(struct inode *inode)	{
	DB *accessdb;
	DBT data, key;
	int *freq_pointer;
	int ret=0, err=0;


	if (itopd(inode)->freq_number == 0)	{
		ret = 1;
		goto out;
	}


	accessdb = stopd(inode->i_sb)->accessdb;
	
	memset(&data, 0, sizeof(DBT));
        memset(&key, 0, sizeof(DBT));


	// the key is the inode number and data is a 8 byte value containing policy bits
	// and the frequency check number
	
	key.data = KMALLOC(sizeof(long int), GFP_KERNEL);

	key.size = sizeof(long int);
	*((long int*)(key.data)) = itohi(inode)->i_ino;

	//  get the number of times the file was accessed, from the accessdb
	
	err = accessdb->get(accessdb, NULL, &key, &data, 0);
        if (err == 0)   {
		freq_pointer = (int*)data.data;
		if (*freq_pointer >= (itopd(inode)->freq_number)-1)	{
			ret = 1;
			*freq_pointer = 0;
		}
		else	{
			ret = 0;
			*freq_pointer = *freq_pointer + 1;
		}
		
		if (accessdb->put(accessdb, NULL, &key, &data, 0))
			printk(KERN_WARNING "Access DB update failed");

	}
	else	{	
		ret = 1;
	}

	KFREE(key.data);

out:
	return ret;
}


/* Wrapper for updating the data checksums */

void update_data_db(DB* datadb, unsigned long ino, unsigned long pno, void* checksum)	{
	add_to_datadb(datadb, ino, pno, checksum);	
}
